#include "main.h"

void cleanup(struct server *server) {
	wlr_backend_destroy(server->backend);
	wlr_allocator_destroy(server->alloc);
	wl_display_destroy(server->wl_display);
}

void rendermon(struct wl_listener *listener, void *data) {
	struct output *output = wl_container_of(listener, output, frame);
	struct server *server = output->server;
	struct wlr_output *wlr_output = data;

	if (!wlr_output_attach_render(output->wlr_output, NULL))
		return;

	wlr_output_render_software_cursors(output->wlr_output, NULL);

	int width, height;
	wlr_output_effective_resolution(output->wlr_output, &width, &height);

	wlr_renderer_begin(server->renderer, width, height);

	float color[4] = {0.3, 0.3, 0.3, 1.0};
	wlr_renderer_clear(server->renderer, color);

	wlr_renderer_end(server->renderer);

	wlr_output_commit(output->wlr_output);
}

void cleanupmon(struct wl_listener *listener, void *data) {
	struct output *output = wl_container_of(listener, output, destroy);

	wl_list_remove(&output->link);
	wl_list_remove(&output->destroy.link);
	wl_list_remove(&output->frame.link);

	free(output);
}

void createmon(struct wl_listener *listener, void *data) {
	struct server *server = wl_container_of(listener, server, new_output);
	struct wlr_output *wlr_output = data;

	struct output *output = malloc(sizeof(struct output));
	if (output == NULL) {
		cleanup(server);
		die("createmon", 1);
	}

	wlr_output_init_render(wlr_output, server->alloc, server->renderer);

	if (wl_list_length(&wlr_output->modes) > 0) {
		wlr_output_set_mode(wlr_output, wlr_output_preferred_mode(wlr_output));
		wlr_output_enable(wlr_output, 1);
		if (!wlr_output_commit(wlr_output))
			return;
	}

	output->server = server;
	output->wlr_output = wlr_output;
	wl_list_insert(&server->outputs, &output->link);

	output->destroy.notify = cleanupmon;
	wl_signal_add(&wlr_output->events.destroy, &output->destroy);

	output->frame.notify = rendermon;
	wl_signal_add(&wlr_output->events.frame, &output->frame);

	wlr_output_commit(wlr_output);
	wlr_output_layout_add_auto(server->output_layout, wlr_output);
}

void server_init(struct server *server) {
	server->wl_display = wl_display_create();
	assert(server->wl_display);

	server->backend = wlr_backend_autocreate(server->wl_display);
	assert(server->backend);

	server->renderer = wlr_renderer_autocreate(server->backend);
	assert(server->renderer);

	wlr_renderer_init_wl_display(server->renderer, server->wl_display);

	server->alloc = wlr_allocator_autocreate(server->backend, server->renderer);
	assert(server->alloc);

	wlr_compositor_create(server->wl_display, server->renderer);
	wlr_data_device_manager_create(server->wl_display);

	server->output_layout = wlr_output_layout_create();

	/* Init monitors */
	wl_list_init(&server->outputs);
	server->new_output.notify = createmon;
	wl_signal_add(&server->backend->events.new_output, &server->new_output);

	const char *socket = wl_display_add_socket_auto(server->wl_display);
	if (!socket) {
		cleanup(server);
		die("wl_display_add_socket_auto", 1);
	}

	printf("Running compositor on wayland display '%s'\n", socket);
	setenv("WAYLAND_DISPLAY", socket, 1);

	if (!wlr_backend_start(server->backend)) {
		cleanup(server);
		die("wlr_backend_start", 1);
	}
}

int main(void) {
	if (!getenv("XDG_RUNTIME_DIR"))
		die("XDG_RUNTIME_DIR must be set", 0);

	int sig[] = {SIGCHLD, SIGINT, SIGTERM, SIGPIPE};
	for (size_t i = 0; i < sizeof(sig) / sizeof(sig[0]); i++)
			signal(sig[i], NULL);

	wlr_log_init(LOG_LEVEL, NULL);

	struct server server;
	server_init(&server);

	wl_display_run(server.wl_display);
	cleanup(&server);
	return 0;
}
